@@ -10,14 +10,16 @@ import (
 	"sync"
 	"time"
 
+	configlistersv1 "github.com/openshift/client-go/config/listers/config/v1"
 	"github.com/prometheus/client_golang/prometheus"
 	"golang.org/x/time/rate"
-	"k8s.io/klog"
-
+	corev1 "k8s.io/api/core/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/util/errors"
 	utilruntime "k8s.io/apimachinery/pkg/util/runtime"
 	"k8s.io/apimachinery/pkg/util/wait"
+	"k8s.io/client-go/tools/record"
+	"k8s.io/klog/v2"
 
 	configv1 "github.com/openshift/api/config/v1"
 
@@ -29,7 +31,7 @@ import (
 
 // ConfigSyncWorker abstracts how the image is synchronized to the server. Introduced for testing.
 type ConfigSyncWorker interface {
-	Start(ctx context.Context, maxWorkers int)
+	Start(ctx context.Context, maxWorkers int, cvoOptrName string, lister configlistersv1.ClusterVersionLister)
 	Update(generation int64, desired configv1.Update, overrides []configv1.ComponentOverride, state payload.State) *SyncWorkerStatus
 	StatusCh() <-chan SyncWorkerStatus
 }
@@ -96,7 +98,7 @@ type SyncWorkerStatus struct {
 
 	LastProgress time.Time
 
-	Actual   configv1.Update
+	Actual   configv1.Release
 	Verified bool
 }
 
@@ -130,7 +132,7 @@ type SyncWorker struct {
 	retriever     PayloadRetriever
 	builder       payload.ResourceBuilder
 	preconditions precondition.List
-	reconciling   bool
+	eventRecorder record.EventRecorder
 
 	// minimumReconcileInterval is the minimum time between reconcile attempts, and is
 	// used to define the maximum backoff interval when syncOnce() returns an error.
@@ -157,11 +159,12 @@ type SyncWorker struct {
 
 // NewSyncWorker initializes a ConfigSyncWorker that will retrieve payloads to disk, apply them via builder
 // to a server, and obey limits about how often to reconcile or retry on errors.
-func NewSyncWorker(retriever PayloadRetriever, builder payload.ResourceBuilder, reconcileInterval time.Duration, backoff wait.Backoff, exclude string) *SyncWorker {
+func NewSyncWorker(retriever PayloadRetriever, builder payload.ResourceBuilder, reconcileInterval time.Duration, backoff wait.Backoff, exclude string, eventRecorder record.EventRecorder) *SyncWorker {
 	return &SyncWorker{
-		retriever: retriever,
-		builder:   builder,
-		backoff:   backoff,
+		retriever:     retriever,
+		builder:       builder,
+		backoff:       backoff,
+		eventRecorder: eventRecorder,
 
 		minimumReconcileInterval: reconcileInterval,
 
@@ -178,8 +181,8 @@ func NewSyncWorker(retriever PayloadRetriever, builder payload.ResourceBuilder,
 // NewSyncWorkerWithPreconditions initializes a ConfigSyncWorker that will retrieve payloads to disk, apply them via builder
 // to a server, and obey limits about how often to reconcile or retry on errors.
 // It allows providing preconditions for loading payload.
-func NewSyncWorkerWithPreconditions(retriever PayloadRetriever, builder payload.ResourceBuilder, preconditions precondition.List, reconcileInterval time.Duration, backoff wait.Backoff, exclude string) *SyncWorker {
-	worker := NewSyncWorker(retriever, builder, reconcileInterval, backoff, exclude)
+func NewSyncWorkerWithPreconditions(retriever PayloadRetriever, builder payload.ResourceBuilder, preconditions precondition.List, reconcileInterval time.Duration, backoff wait.Backoff, exclude string, eventRecorder record.EventRecorder) *SyncWorker {
+	worker := NewSyncWorker(retriever, builder, reconcileInterval, backoff, exclude, eventRecorder)
 	worker.preconditions = preconditions
 	return worker
 }
@@ -210,7 +213,13 @@ func (w *SyncWorker) Update(generation int64, desired configv1.Update, overrides
 		w.work.Generation = generation
 	}
 
-	if work.Empty() || equalSyncWork(w.work, work) {
+	if work.Empty() {
+		klog.V(5).Info("Update work has no release image; ignoring requested change")
+		return w.status.DeepCopy()
+	}
+
+	if equalSyncWork(w.work, work) {
+		klog.V(5).Info("Update work is equal to current target; no change required")
 		return w.status.DeepCopy()
 	}
 
@@ -221,19 +230,25 @@ func (w *SyncWorker) Update(generation int64, desired configv1.Update, overrides
 		w.status = SyncWorkerStatus{
 			Generation:  generation,
 			Reconciling: state.Reconciling(),
-			Actual:      work.Desired,
+			Actual: configv1.Release{
+				Version: work.Desired.Version,
+				Image:   work.Desired.Image,
+			},
 		}
 	}
 
 	// notify the sync loop that we changed config
 	w.work = work
 	if w.cancelFn != nil {
+		klog.V(5).Info("Cancel the sync worker's current loop")
 		w.cancelFn()
 		w.cancelFn = nil
 	}
 	select {
 	case w.notify <- struct{}{}:
+		klog.V(5).Info("Notify the sync worker that new work is available")
 	default:
+		klog.V(5).Info("The sync worker has already been notified that new work is available")
 	}
 
 	return w.status.DeepCopy()
@@ -242,7 +257,7 @@ func (w *SyncWorker) Update(generation int64, desired configv1.Update, overrides
 // Start periodically invokes run, detecting whether content has changed.
 // It is edge-triggered when Update() is invoked and level-driven after the
 // syncOnce() has succeeded for a given input (we are said to be "reconciling").
-func (w *SyncWorker) Start(ctx context.Context, maxWorkers int) {
+func (w *SyncWorker) Start(ctx context.Context, maxWorkers int, cvoOptrName string, lister configlistersv1.ClusterVersionLister) {
 	klog.V(5).Infof("Starting sync worker")
 
 	work := &SyncWork{}
@@ -307,11 +322,15 @@ func (w *SyncWorker) Start(ctx context.Context, maxWorkers int) {
 				w.lock.Unlock()
 				defer cancelFn()
 
+				config, err := lister.Get(cvoOptrName)
+				if err != nil {
+					return err
+				}
 				// reporter hides status updates that occur earlier than the previous failure,
 				// so that we don't fail, then immediately start reporting an earlier status
 				reporter := &statusWrapper{w: w, previousStatus: w.Status()}
 				klog.V(5).Infof("Previous sync status: %#v", reporter.previousStatus)
-				return w.syncOnce(ctx, work, maxWorkers, reporter)
+				return w.syncOnce(ctx, work, maxWorkers, reporter, config)
 			}()
 			if err != nil {
 				// backoff wait
@@ -329,7 +348,9 @@ func (w *SyncWorker) Start(ctx context.Context, maxWorkers int) {
 				utilruntime.HandleError(fmt.Errorf("unable to synchronize image (waiting %s): %v", interval, err))
 				continue
 			}
-			klog.V(5).Infof("Sync succeeded, reconciling")
+			if work.State != payload.ReconcilingPayload {
+				klog.V(4).Infof("Sync succeeded, transitioning from %s to %s", work.State, payload.ReconcilingPayload)
+			}
 
 			work.Completed++
 			work.State = payload.ReconcilingPayload
@@ -350,14 +371,14 @@ type statusWrapper struct {
 func (w *statusWrapper) Report(status SyncWorkerStatus) {
 	p := w.previousStatus
 	if p.Failure != nil && status.Failure == nil {
-		if p.Actual == status.Actual {
+		if p.Actual.Image == status.Actual.Image {
 			if status.Fraction < p.Fraction {
 				klog.V(5).Infof("Dropping status report from earlier in sync loop")
 				return
 			}
 		}
 	}
-	if status.Fraction > p.Fraction || status.Completed > p.Completed || (status.Failure == nil && status.Actual != p.Actual) {
+	if status.Fraction > p.Fraction || status.Completed > p.Completed || (status.Failure == nil && status.Actual.Image != p.Actual.Image) {
 		status.LastProgress = time.Now()
 	}
 	if status.Generation == 0 {
@@ -382,10 +403,8 @@ func (w *SyncWorker) calculateNext(work *SyncWork) bool {
 	// the state Update() calculated (to allow us to start in reconciling)
 	if work.Empty() {
 		work.State = w.work.State
-	} else {
-		if changed {
-			work.State = payload.UpdatingPayload
-		}
+	} else if changed {
+		work.State = payload.UpdatingPayload
 	}
 	// always clear the completed variable if we are not reconciling
 	if work.State != payload.ReconcilingPayload {
@@ -436,12 +455,12 @@ func (w *SyncWorker) updateStatus(update SyncWorkerStatus) {
 	w.lock.Lock()
 	defer w.lock.Unlock()
 
-	klog.V(5).Infof("Status change %#v", update)
+	klog.V(6).Infof("Status change %#v", update)
 	w.status = update
 	select {
 	case w.report <- update:
 	default:
-		if klog.V(5) {
+		if klog.V(6).Enabled() {
 			klog.Infof("Status report channel was full %#v", update)
 		}
 	}
@@ -467,50 +486,82 @@ func (w *SyncWorker) Status() *SyncWorkerStatus {
 // sync retrieves the image and applies it to the server, returning an error if
 // the update could not be completely applied. The status is updated as we progress.
 // Cancelling the context will abort the execution of the sync.
-func (w *SyncWorker) syncOnce(ctx context.Context, work *SyncWork, maxWorkers int, reporter StatusReporter) error {
-	klog.V(4).Infof("Running sync %s (force=%t) on generation %d in state %s at attempt %d", versionString(work.Desired), work.Desired.Force, work.Generation, work.State, work.Attempt)
-	update := work.Desired
+func (w *SyncWorker) syncOnce(ctx context.Context, work *SyncWork, maxWorkers int, reporter StatusReporter, clusterVersion *configv1.ClusterVersion) error {
+	desired := configv1.Release{
+		Version: work.Desired.Version,
+		Image:   work.Desired.Image,
+	}
+	klog.V(4).Infof("Running sync %s (force=%t) on generation %d in state %s at attempt %d", versionString(desired), work.Desired.Force, work.Generation, work.State, work.Attempt)
 
 	// cache the payload until the release image changes
 	validPayload := w.payload
-	if validPayload == nil || !equalUpdate(configv1.Update{Image: validPayload.ReleaseImage}, configv1.Update{Image: update.Image}) {
+	if validPayload != nil && validPayload.Release.Image == desired.Image {
+		// possibly complain here if Version, etc. diverges from the payload content
+		desired = validPayload.Release
+	} else if validPayload == nil || !equalUpdate(configv1.Update{Image: validPayload.Release.Image}, configv1.Update{Image: desired.Image}) {
 		klog.V(4).Infof("Loading payload")
+		cvoObjectRef := &corev1.ObjectReference{APIVersion: "config.openshift.io/v1", Kind: "ClusterVersion", Name: "version", Namespace: "openshift-cluster-version"}
+		w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeNormal, "RetrievePayload", "retrieving payload version=%q image=%q", desired.Version, desired.Image)
 		reporter.Report(SyncWorkerStatus{
 			Generation:  work.Generation,
 			Step:        "RetrievePayload",
 			Initial:     work.State.Initializing(),
 			Reconciling: work.State.Reconciling(),
-			Actual:      update,
+			Actual:      desired,
 		})
-		info, err := w.retriever.RetrievePayload(ctx, update)
+		info, err := w.retriever.RetrievePayload(ctx, work.Desired)
 		if err != nil {
+			w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeWarning, "RetrievePayloadFailed", "retrieving payload failed version=%q image=%q failure=%v", desired.Version, desired.Image, err)
 			reporter.Report(SyncWorkerStatus{
 				Generation:  work.Generation,
 				Failure:     err,
 				Step:        "RetrievePayload",
 				Initial:     work.State.Initializing(),
 				Reconciling: work.State.Reconciling(),
-				Actual:      update,
+				Actual:      desired,
 			})
 			return err
 		}
 
-		payloadUpdate, err := payload.LoadUpdate(info.Directory, update.Image, w.exclude)
+		w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeNormal, "VerifyPayload", "verifying payload version=%q image=%q", desired.Version, desired.Image)
+		payloadUpdate, err := payload.LoadUpdate(info.Directory, desired.Image, w.exclude)
 		if err != nil {
+			w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeWarning, "VerifyPayloadFailed", "verifying payload failed version=%q image=%q failure=%v", desired.Version, desired.Image, err)
 			reporter.Report(SyncWorkerStatus{
 				Generation:  work.Generation,
 				Failure:     err,
 				Step:        "VerifyPayload",
 				Initial:     work.State.Initializing(),
 				Reconciling: work.State.Reconciling(),
-				Actual:      update,
+				Actual:      desired,
 				Verified:    info.Verified,
 			})
 			return err
 		}
+
 		payloadUpdate.VerifiedImage = info.Verified
 		payloadUpdate.LoadedAt = time.Now()
 
+		if work.Desired.Version == "" {
+			work.Desired.Version = payloadUpdate.Release.Version
+			desired.Version = payloadUpdate.Release.Version
+		} else if payloadUpdate.Release.Version != work.Desired.Version {
+			err = fmt.Errorf("release image version %s does not match the expected upstream version %s", payloadUpdate.Release.Version, work.Desired.Version)
+			w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeWarning, "VerifyPayloadVersionFailed", "verifying payload failed version=%q image=%q failure=%v", work.Desired.Version, work.Desired.Image, err)
+			/* FIXME: Ignore for now.  I will make this fatal in a follow-up pivot
+			reporter.Report(SyncWorkerStatus{
+				Generation:  work.Generation,
+				Failure:     err,
+				Step:        "VerifyPayloadVersion",
+				Initial:     work.State.Initializing(),
+				Reconciling: work.State.Reconciling(),
+				Actual:      desired,
+				Verified:    info.Verified,
+			})
+			return err
+			*/
+		}
+
 		// need to make sure the payload is only set when the preconditions have been successful
 		if len(w.preconditions) == 0 {
 			klog.V(4).Info("No preconditions configured.")
@@ -522,29 +573,33 @@ func (w *SyncWorker) syncOnce(ctx context.Context, work *SyncWork, maxWorkers in
 				Step:        "PreconditionChecks",
 				Initial:     work.State.Initializing(),
 				Reconciling: work.State.Reconciling(),
-				Actual:      update,
+				Actual:      desired,
 				Verified:    info.Verified,
 			})
-			if err := precondition.Summarize(w.preconditions.RunAll(ctx, precondition.ReleaseContext{DesiredVersion: payloadUpdate.ReleaseVersion})); err != nil {
-				if update.Force {
+			if err := precondition.Summarize(w.preconditions.RunAll(ctx, precondition.ReleaseContext{DesiredVersion: payloadUpdate.Release.Version}, clusterVersion)); err != nil {
+				if work.Desired.Force {
 					klog.V(4).Infof("Forcing past precondition failures: %s", err)
+					w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeWarning, "PreconditionsForced", "preconditions forced for payload loaded version=%q image=%q failures=%v", desired.Version, desired.Image, err)
 				} else {
+					w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeWarning, "PreconditionsFailed", "preconditions failed for payload loaded version=%q image=%q failures=%v", desired.Version, desired.Image, err)
 					reporter.Report(SyncWorkerStatus{
 						Generation:  work.Generation,
 						Failure:     err,
 						Step:        "PreconditionChecks",
 						Initial:     work.State.Initializing(),
 						Reconciling: work.State.Reconciling(),
-						Actual:      update,
+						Actual:      desired,
 						Verified:    info.Verified,
 					})
 					return err
 				}
 			}
+			w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeNormal, "PreconditionsPassed", "preconditions passed for payload loaded version=%q image=%q", desired.Version, desired.Image)
 		}
 
 		w.payload = payloadUpdate
-		klog.V(4).Infof("Payload loaded from %s with hash %s", payloadUpdate.ReleaseImage, payloadUpdate.ManifestHash)
+		w.eventRecorder.Eventf(cvoObjectRef, corev1.EventTypeNormal, "PayloadLoaded", "payload loaded version=%q image=%q", desired.Version, desired.Image)
+		klog.V(4).Infof("Payload loaded from %s with hash %s", desired.Image, payloadUpdate.ManifestHash)
 	}
 
 	return w.apply(ctx, w.payload, work, maxWorkers, reporter)
@@ -554,14 +609,7 @@ func (w *SyncWorker) syncOnce(ctx context.Context, work *SyncWork, maxWorkers in
 // Cancelling the context will abort the execution of the sync. Will be executed in parallel if
 // maxWorkers is set greater than 1.
 func (w *SyncWorker) apply(ctx context.Context, payloadUpdate *payload.Update, work *SyncWork, maxWorkers int, reporter StatusReporter) error {
-	update := configv1.Update{
-		Version: payloadUpdate.ReleaseVersion,
-		Image:   payloadUpdate.ReleaseImage,
-		Force:   work.Desired.Force,
-	}
-
 	// encapsulate status reporting in a threadsafe updater
-	version := payloadUpdate.ReleaseVersion
 	total := len(payloadUpdate.Manifests)
 	cr := &consistentReporter{
 		status: SyncWorkerStatus{
@@ -569,11 +617,11 @@ func (w *SyncWorker) apply(ctx context.Context, payloadUpdate *payload.Update, w
 			Initial:     work.State.Initializing(),
 			Reconciling: work.State.Reconciling(),
 			VersionHash: payloadUpdate.ManifestHash,
-			Actual:      update,
+			Actual:      payloadUpdate.Release,
 			Verified:    payloadUpdate.VerifiedImage,
 		},
 		completed: work.Completed,
-		version:   version,
+		version:   payloadUpdate.Release.Version,
 		total:     total,
 		reporter:  reporter,
 	}
@@ -651,7 +699,6 @@ func (w *SyncWorker) apply(ctx context.Context, payloadUpdate *payload.Update, w
 			cr.Update()
 
 			klog.V(4).Infof("Running sync for %s", task)
-			klog.V(5).Infof("Manifest: %s", string(task.Manifest.Raw))
 
 			ov, ok := getOverrideForManifest(work.Overrides, task.Manifest)
 			if ok && ov.Unmanaged {
@@ -659,7 +706,7 @@ func (w *SyncWorker) apply(ctx context.Context, payloadUpdate *payload.Update, w
 				continue
 			}
 
-			if err := task.Run(ctx, version, w.builder, work.State); err != nil {
+			if err := task.Run(ctx, payloadUpdate.Release.Version, w.builder, work.State); err != nil {
 				return err
 			}
 			cr.Inc()
@@ -814,7 +861,7 @@ func summarizeTaskGraphErrors(errs []error) error {
 	}
 
 	// log the errors to assist in debugging future summarization
-	if klog.V(4) {
+	if klog.V(4).Enabled() {
 		klog.Infof("Summarizing %d errors", len(errs))
 		for _, err := range errs {
 			if uErr, ok := err.(*payload.UpdateError); ok {
@@ -972,21 +1019,20 @@ func ownerRefModifier(config *configv1.ClusterVersion) resourcebuilder.MetaV1Obj
 
 // runThrottledStatusNotifier invokes fn every time ch is updated, but no more often than once
 // every interval. If bucket is non-zero then the channel is throttled like a rate limiter bucket.
-func runThrottledStatusNotifier(stopCh <-chan struct{}, interval time.Duration, bucket int, ch <-chan SyncWorkerStatus, fn func()) {
+func runThrottledStatusNotifier(ctx context.Context, interval time.Duration, bucket int, ch <-chan SyncWorkerStatus, fn func()) {
 	// notify the status change function fairly infrequently to avoid updating
 	// the caller status more frequently than is needed
 	throttle := rate.NewLimiter(rate.Every(interval), bucket)
-	wait.Until(func() {
-		ctx := context.Background()
+	wait.UntilWithContext(ctx, func(ctx context.Context) {
 		var last SyncWorkerStatus
 		for {
 			select {
-			case <-stopCh:
+			case <-ctx.Done():
 				return
 			case next := <-ch:
 				// only throttle if we aren't on an edge
-				if next.Generation == last.Generation && next.Actual == last.Actual && next.Reconciling == last.Reconciling && (next.Failure != nil) == (last.Failure != nil) {
-					if err := throttle.Wait(ctx); err != nil {
+				if next.Generation == last.Generation && next.Actual.Image == last.Actual.Image && next.Reconciling == last.Reconciling && (next.Failure != nil) == (last.Failure != nil) {
+					if err := throttle.Wait(ctx); err != nil && err != context.Canceled && err != context.DeadlineExceeded {
 						utilruntime.HandleError(fmt.Errorf("unable to throttle status notification: %v", err))
 					}
 				}
@@ -995,5 +1041,5 @@ func runThrottledStatusNotifier(stopCh <-chan struct{}, interval time.Duration,
 				fn()
 			}
 		}
-	}, 1*time.Second, stopCh)
+	}, 1*time.Second)
 }
